/*
 * Copyright 2020 Confluent Inc.
 *
 * Licensed under the Confluent Community License (the "License"); you may not use
 * this file except in compliance with the License.  You may obtain a copy of the
 * License at
 *
 * http://www.confluent.io/confluent-community-license
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 */

package io.confluent.kafkarest.exceptions.v2;

import io.confluent.kafkarest.exceptions.StatusCodeException;
import javax.ws.rs.core.Response;
import javax.ws.rs.ext.ExceptionMapper;

public final class V2ExceptionMapper implements ExceptionMapper<StatusCodeException> {

  @Override
  public Response toResponse(StatusCodeException exception) {
    return Response.status(exception.getStatus())
        .entity(
            ErrorResponse.create(
                exception.getCode(),
                String.format("%s: %s", exception.getTitle(), exception.getDetail())))
        .build();
  }
}
